package org.seasar.cayenne.dao.impl;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.seasar.cayenne.S2DataContextFactory;
import org.seasar.cayenne.dao.CayenneDaoCommand;
import org.seasar.cayenne.dao.CayenneDaoMetaData;
import org.seasar.framework.beans.BeanDesc;
import org.seasar.framework.beans.MethodNotFoundRuntimeException;
import org.seasar.framework.beans.factory.BeanDescFactory;
import org.seasar.framework.util.FieldUtil;
import org.seasar.framework.util.MethodUtil;
import org.seasar.framework.util.StringUtil;

/**
 * @author Takahiro MATSUNAGA
 * @see CayenneDaoMetaData
 */
public class CayenneDaoMetaDataImpl implements CayenneDaoMetaData {

    private static final String[] INSERT_NAMES = {
        "insert", "add", "create",
    };

    private static final String[] UPDATE_NAMES = {
        "update", "modify", "store",
    };

    private static final String[] DELETE_NAMES = {
        "delete", "remove",
    };

    private S2DataContextFactory S2DataContextFactory_;

    private Class daoClass_;

    private BeanDesc daoBeanDesc_;

    private Class beanClass_;

    private Map cayenneCommands_ = new HashMap();

    /**
     * @param S2DataContextFactory
     */
    public CayenneDaoMetaDataImpl(S2DataContextFactory S2DataContextFactory, Class daoClass) {
        S2DataContextFactory_ = S2DataContextFactory;

        daoClass_ = daoClass;
        daoBeanDesc_ = BeanDescFactory.getBeanDesc(daoClass);
        Field beanField = daoBeanDesc_.getField(BEAN_KEY);
        beanClass_ = (Class) FieldUtil.get(beanField, null);

        setupCayenneCommand();
    }

    private void setupCayenneCommand() {
        String[] names = daoBeanDesc_.getMethodNames();
        for (int i = 0; i < names.length; i++) {
            Method[] methods = daoBeanDesc_.getMethods(names[i]);
            if (methods.length == 1 && MethodUtil.isAbstract(methods[0])) {
                setupMethod(methods[0]);
            }
        }
    }

    private void setupMethod(Method method) {
        String methodName = method.getName();

        if (isMethod(INSERT_NAMES, methodName)) {
            setupInsertCommand(method);
        } else if (isMethod(UPDATE_NAMES, methodName)) {
            setupUpdateCommand(method);
        } else if (isMethod(DELETE_NAMES, methodName)) {
            setupDeleteCommand(method);
        } else {
            setupSelectMethod(method);
        }
    }

    private void setupSelectMethod(Method method) {
        String query = getQuery(method.getName());
        String[] argNames = getArgNames(method.getName());

        AbstractSelectCommand cmd = null;
        if (query != null && !startsWithOrderBy(query)) {
            argNames = getArgNamesByQuery(query);

            cmd = new SelectQueryCommand(S2DataContextFactory_, beanClass_);
        } else {
            if (argNames.length == 0 && method.getParameterTypes().length == 1) {
                cmd = new SelectDtoCommand(S2DataContextFactory_, beanClass_);
            } else {
                cmd = new SelectAutoCommand(S2DataContextFactory_, beanClass_);
            }
        }
        cmd.setArgNames(argNames);
        cmd.setMethod(method);
        cmd.setQuery(query);

        cayenneCommands_.put(method.getName(), cmd);
    }

    private boolean isMethod(String[] methodNameList, String methodName) {
        for (int i = 0; i < methodNameList.length; i++) {
            if (methodName.startsWith(methodNameList[i])) {
                return true;
            }
        }
        return false;
    }

    private String[] getArgNames(String methodName) {
        String argsKey = methodName + ARGS_KEY_SUFFIX;

        if (daoBeanDesc_.hasField(argsKey)) {
            Field argNamesField = daoBeanDesc_.getField(argsKey);
            String argNames = (String) FieldUtil.get(argNamesField, null);
            return StringUtil.split(argNames, " ,");
        } else {
            return new String[0];
        }
    }

    private String[] getArgNamesByQuery(String query) {
        List args = new ArrayList();

        String[] querys = query.trim().split(" ");
        for (int i = 0; i < querys.length; i++) {
            if (querys[i].indexOf("$") >= 0) {
                int st = querys[i].indexOf("$");
                int ed = querys[i].indexOf(")");

                if (st >= 0) {
                    if (ed < 0) {
                        args.add(querys[i].substring(st + 1));
                    } else {
                        args.add(querys[i].substring(st + 1, ed));
                    }
                }
            }
        }

        return (String[]) args.toArray(new String[0]);
    }

    private String getQuery(String methodName) {
        String queryKey = methodName + QUERY_KEY_SUFFIX;

        if (daoBeanDesc_.hasField(queryKey)) {
            Field queryNamesField = daoBeanDesc_.getField(queryKey);
            return (String) FieldUtil.get(queryNamesField, null);
        } else {
            return null;
        }
    }

    private static boolean startsWithOrderBy(String query) {
        return StringUtil.startsWith(query, "order by");
    }

    private void setupInsertCommand(Method method) {
        InsertCommand cmd = new InsertCommand(S2DataContextFactory_, beanClass_);

        cayenneCommands_.put(method.getName(), cmd);
    }

    private void setupUpdateCommand(Method method) {
        UpdateCommand cmd = new UpdateCommand(S2DataContextFactory_, beanClass_);

        cayenneCommands_.put(method.getName(), cmd);
    }

    private void setupDeleteCommand(Method method) {
        DeleteCommand cmd = new DeleteCommand(S2DataContextFactory_, beanClass_);
 
        cayenneCommands_.put(method.getName(), cmd);
    }

    public CayenneDaoCommand getDaoCommand(String methodName) throws MethodNotFoundRuntimeException {
        return (CayenneDaoCommand) cayenneCommands_.get(methodName);
    }

}
